
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#include <iostream>

#include "Character.hh"
#include "Log.hh"

	
struct SpringForce
{
	explicit SpringForce(Mf::Vector2 x) :
		location(x) {}

	const Mf::Vector2& operator () (const Mf::LinearState<2>& state)
	{
		Mf::Vector2 x = state.position - location;
		Mf::Scalar mag = x.length();
		Mf::Scalar d = 50.0;

		// spring:
		//current.force += -15.0 * x - 1.5 * current.velocity;
		force = -20.0 * (mag - d) * (x / mag) - 2.0 * state.velocity;

		return force;
	}

private:

	Mf::Vector2 force;
	Mf::Vector2 location;
};

struct ResistanceForce
{
	explicit ResistanceForce(Mf::Scalar scale = 1.0) :
		k(scale) {}

	const Mf::Vector2& operator () (const Mf::LinearState<2>& state)
	{
		force = -k * state.velocity;
		return force;
	}

private:

	Mf::Vector2	force;
	Mf::Scalar	k;
};


Character::Character(const std::string& name) :
	tilemap(name),
	animation(name)
{
	current.init();

	current.mass = 1.0;
	current.inverseMass = 1.0 / current.mass;

	// forces
	current.force = Mf::Vector2(0.0, 0.0);
	current.forces.push_back(SpringForce(Mf::Vector2(500.0, 200.0)));
	current.forces.push_back(ResistanceForce(2.0));
	current.forces.push_back(Mf::LinearState<2>::GravityForce(-1000.0));

	// starting position
	current.position = Mf::Vector2(64.0, 64.0);
	current.momentum = Mf::Vector2(0.0, 0.0);
	current.recalculate();

	previous = current;
}


void Character::update(Mf::Scalar t, Mf::Scalar dt)
{
	previous = current;

	//Mf::Vector2 x = current.position - Mf::Vector2(500.0, 200.0);
	//Mf::Scalar mag = x.length();
	//Mf::Scalar d = 50.0;

	//// gravity:
	//current.force = Mf::Vector2(0.0, -2000.0);
	//// spring:
	////current.force += -15.0 * x - 1.5 * current.velocity;
	//current.force += -20.0 * (mag - d) * (x / mag) - 2.0 * current.velocity;
	//// internal:
	//current.force += userForce;
	//current.recalculate();
	//std::cout << "force: " << current.momentum << std::endl;

	//Mf::euler<State,Derivative>(current, t, dt);

	current.integrate(t, dt);
	animation.update(t, dt);

	Mf::Vector3 center(current.position[0], current.position[1], z);
	Mf::Vector3 a(current.position[0] - 16.0, current.position[1] - 16.0, z);
	Mf::Vector3 b(current.position[0] + 16.0, current.position[1] + 16.0, z);

	aabb_.init(a, b);
	sphere_.init(center, a);
}


void Character::draw(Mf::Scalar alpha) const
{
	Mf::Vector2 position = cml::lerp(previous.position, current.position, alpha);

	//glColor3f(1.0f, 1.0f, 1.0f);
	tilemap.bind();

	Tilemap::Index frame = animation.getFrame();

	Tilemap::Orientation orientation = Tilemap::NORMAL;

	if (current.velocity[0] < 0.0) orientation = Tilemap::REVERSE;

	Mf::Scalar coords[8];
	tilemap.getTileCoords(frame, coords, orientation);

	Mf::Scalar s = 16.0;

	glBegin(GL_TRIANGLE_FAN);
		glTexCoord2f(coords[0], coords[1]);
		glVertex3(position[0]-s, position[1]-s, z);
		glTexCoord2f(coords[2], coords[3]);
		glVertex3(position[0]+s, position[1]-s, z);
		glTexCoord2f(coords[4], coords[5]);
		glVertex3(position[0]+s, position[1]+s, z);
		glTexCoord2f(coords[6], coords[7]);
		glVertex3(position[0]-s, position[1]+s, z);
	glEnd();

	//glColor3f(0.0f, 0.0f, 0.0f);
	Mf::Texture::resetBind();

	glBegin(GL_TRIANGLES);
		glVertex3(480.0, 190.0, 64.0);
		glVertex3(520.0, 190.0, 64.0);
		glVertex3(500.0, 210.0, 64.0);
	glEnd();

	//glColor3f(1.0f, 1.0f, 1.0f);
}


bool Character::isInsideAabb(const Mf::Aabb& aabb) const
{
	// make sure the entity is fully inside the volume
	if (!(aabb_.max[0] < aabb.max[0] &&
		  aabb_.min[0] > aabb.min[0] &&
		  aabb_.max[1] < aabb.max[1] &&
		  aabb_.min[1] > aabb.min[1] &&
		  aabb_.max[2] < aabb.max[2] &&
		  aabb_.min[2] > aabb.min[2]))
	{
		return false;
	}

	return true;
}

int Character::getOctant(const Mf::Aabb& aabb) const
{
	int octantNum = -1;

	Mf::Plane::Halfspace halfspace;

	Mf::Plane xy = aabb.getPlaneXY();
	halfspace = xy.intersects(sphere_);
	if (halfspace == Mf::Plane::INTERSECT)
	{
		halfspace = xy.intersects(aabb_);
	}

	if (halfspace == Mf::Plane::POSITIVE)
	{
		Mf::Plane xz = aabb.getPlaneXZ();
		halfspace = xz.intersects(sphere_);
		if (halfspace == Mf::Plane::INTERSECT)
		{
			halfspace = xz.intersects(aabb_);
		}

		if (halfspace == Mf::Plane::POSITIVE)
		{
			Mf::Plane yz = aabb.getPlaneYZ();
			halfspace = yz.intersects(sphere_);
			if (halfspace == Mf::Plane::INTERSECT)
			{
				halfspace = yz.intersects(aabb_);
			}

			if (halfspace == Mf::Plane::POSITIVE)
			{
				octantNum = 2;
			}
			else if (halfspace == Mf::Plane::NEGATIVE)
			{
				octantNum = 3;
			}
		}
		else if (halfspace == Mf::Plane::NEGATIVE)
		{
			Mf::Plane yz = aabb.getPlaneYZ();
			halfspace = yz.intersects(sphere_);
			if (halfspace == Mf::Plane::INTERSECT)
			{
				halfspace = yz.intersects(aabb_);
			}

			if (halfspace == Mf::Plane::POSITIVE)
			{
				octantNum = 1;
			}
			else if (halfspace == Mf::Plane::NEGATIVE)
			{
				octantNum = 0;
			}
		}
	}
	else if (halfspace == Mf::Plane::NEGATIVE)
	{
		Mf::Plane xz = aabb.getPlaneXZ();
		halfspace = xz.intersects(sphere_);
		if (halfspace == Mf::Plane::INTERSECT)
		{
			halfspace = xz.intersects(aabb_);
		}

		if (halfspace == Mf::Plane::POSITIVE)
		{
			Mf::Plane yz = aabb.getPlaneYZ();
			halfspace = yz.intersects(sphere_);
			if (halfspace == Mf::Plane::INTERSECT)
			{
				halfspace = yz.intersects(aabb_);
			}

			if (halfspace == Mf::Plane::POSITIVE)
			{
				octantNum = 6;
			}
			else if (halfspace == Mf::Plane::NEGATIVE)
			{
				octantNum = 7;
			}
		}
		else if (halfspace == Mf::Plane::NEGATIVE)
		{
			Mf::Plane yz = aabb.getPlaneYZ();
			halfspace = yz.intersects(sphere_);
			if (halfspace == Mf::Plane::INTERSECT)
			{
				halfspace = yz.intersects(aabb_);
			}

			if (halfspace == Mf::Plane::POSITIVE)
			{
				octantNum = 5;
			}
			else if (halfspace == Mf::Plane::NEGATIVE)
			{
				octantNum = 4;
			}
		}
	}

	return octantNum;
}


/** vim: set ts=4 sw=4 tw=80: *************************************************/

